// Graph of strongly connected components -*- c++ -*-

#ifndef COMPONENTGRAPH_H_
# define COMPONENTGRAPH_H_
# ifdef __GNUC__
#  pragma interface
# endif // __GNUC__
# include <cstdio>
# include "typedefs.h"

/** @file ComponentGraph.h
 * Graph of strongly connected components
 */

/** Graph of strongly connected components */
class ComponentGraph
{
public:
  /** Constructor
   * @param graph	the underlying graph
   */
  explicit ComponentGraph (const class Graph& graph);
private:
  /** Copy constructor */
  ComponentGraph (const class ComponentGraph& old);
  /** Assignment operator */
  class ComponentGraph& operator= (const class ComponentGraph& old);
public:
  /** Destructor */
  ~ComponentGraph ();

private:
  /** Open the component graph files with temporary names
   * @return		true if everything succeeded
   */
  inline bool openFiles ();
  /** Set the component number of a state
   * @param state	the state
   * @param comp	the component
   */
  inline void setComponent (card_t state, card_t comp);
  /** Compute the arcs of the component graph */
  inline void computeArcs ();
public:

  /** Compute the strongly connected components
   * @param state	start state of the search
   * @param cond	condition for exploring states (NULL=true)
   * @return		number of strongly connected components (0=fail)
   */
  unsigned compute (card_t state,
		    const class Expression* cond);

  /** Determine the number of strongly connected components in the graph */
  unsigned size () const { return myNumComponents; }

  /** Determine the component a state belongs to
   * @param state	the state number
   * @return		the corresponding component, or CARD_T_MAX if none
   */
  card_t getComponent (card_t state) const;
  /** Determine the states a component consists of
   * @param comp	the component number
   * @return		the amount and numbers of the contained states
   */
  card_t* getStates (card_t comp) const;
  /** Determine the number of successors of a component
   * @param comp	the component number
   * @return		the number of successors
   */
  card_t getNumSucc (card_t comp) const;
  /** Determine the number of predecessors of a component
   * @param comp	the component number
   * @return		the number of predecessors
   */
  card_t getNumPred (card_t comp) const;
  /** Determine the successors of a component
   * @param comp	the component number
   * @return		the amount and numbers of the successor components
   */
  card_t* getSucc (card_t comp) const;
  /** Determine the predecessors of a component
   * @param comp	the component number
   * @return		the amount and numbers of the predecessor components
   */
  card_t* getPred (card_t comp) const;

  /** Determine the shortest path from a state to a component
   * @param state	the state number
   * @param comp	the component number
   * @param pathc	condition that must hold in every state on the path
   * @return		the amount and numbers of the states on the path
   */
  card_t* path (card_t state, card_t comp,
		const class Expression* pathc) const;

  /** Determine the shortest path to a state from a component
   * @param state	the state number
   * @param comp	the component number
   * @param pathc	condition that must hold in every state on the path
   * @return		the amount and numbers of the states on the path
   */
  card_t* rpath (card_t state, card_t comp,
		 const class Expression* pathc) const;

private:
  /** The underlying graph */
  const class Graph& myGraph;
  /** Number of components in the graph (0=not computed) */
  card_t myNumComponents;
  /** Mapping from state numbers to component numbers */
  FILE* myStateComponents;
  /** Component directory */
  FILE* myComponentDirectory;
  /** Mapping from component numbers to state numbers */
  FILE* myComponentStates;
  /** Successor components of a component (forward arcs) */
  FILE* mySuccComponents;
  /** Predecessor components of a component (backward arcs) */
  FILE* myPredComponents;
};

#endif // COMPONENTGRAPH_H_
