// Visualization interface to Graphviz dotty(1) -*- c++ -*-

#ifdef __WIN32
# undef __STRICT_ANSI__
# include <windows.h>
# include <fcntl.h>
# include <process.h>
# undef pipe
# define pipe(fds) _pipe (fds, 0, _O_TEXT)
#endif // __WIN32

#ifdef __GNUC__
# pragma implementation
#endif // __GNUC__
#include "Dotty.h"

#include <cstring>
#include <unistd.h>

#ifdef __alpha
# include </usr/include/stdio.h>
#endif // __alpha

#include "Graph.h"
#include "ComponentGraph.h"
#include "Printer.h"
#include "GlobalMarking.h"
#include "Net.h"
#include "Place.h"
#include "Transition.h"
#include "Arc.h"
#include "Valuation.h"
#include "VariableDefinition.h"
#include "Marking.h"

/** @file Dotty.C
 * Visualization interface to Graphviz dotty(1)
 */

#if defined __WIN32
/** Visualization process handle */
volatile HANDLE vispid;
#elif defined __unix
/** Visualization process identifier */
volatile pid_t vispid;
#endif // __unix
/** Visualization command stream */
FILE* volatile visf;
/** Visualization command file descriptor */
volatile int visfd;
/** Visualization feedback stream */
static FILE* graphf;

class Dotty Dotty::theDotty;

Dotty::Dotty () :
  myGraph (0), myComponentGraph (0),
#ifdef EXPR_COMPILE
  myCompilation (0),
#endif // EXPR_COMPILE
  myOut (0), myPrinter (0)
{
}

Dotty::~Dotty ()
{
  stopVisual ();
  clearVisual ();
}

class Dotty*
Dotty::startVisual (const char* name)
{
  if (visfd)
    return &theDotty;
  theDotty.clearVisual ();
  assert (!visf && !graphf && !vispid && !theDotty.myOut);
  /** pipes for the graph output and the command input */
  int pipe_graph[2], pipe_command[2];
  if (pipe (pipe_graph)) {
    perror ("pipe");
    return 0;
  }
  if (pipe (pipe_command)) {
    perror ("pipe");
    close (pipe_graph[0]), close (pipe_graph[1]);
    return 0;
  }
#if defined __WIN32
  /*
   * Since Windows does fork(2)/exec(2) in single pass, we have to
   * duplicate the file descriptors before executing the child process.
   */
  int old_in = _dup (STDIN_FILENO), old_out = _dup (STDOUT_FILENO);
  _dup2 (pipe_graph[0], STDIN_FILENO);
  _dup2 (pipe_command[1], STDOUT_FILENO);
  /* Create the child process */
  HANDLE pid = reinterpret_cast<HANDLE>(_spawnlp (_P_NOWAIT, name, name, 0));

  _dup2 (old_in, STDIN_FILENO), _dup2 (old_out, STDOUT_FILENO);
  close (old_in), close (old_out);

  if (!pid || pid == INVALID_HANDLE_VALUE) {
    perror (name);
    close (pipe_graph[0]), close (pipe_graph[1]);
    close (pipe_command[0]), close (pipe_command[1]);
    return 0;
  }
#elif defined __unix
  pid_t pid = fork ();
  if (!pid) {
    /* child process */
    dup2 (pipe_graph[0], STDIN_FILENO);
    dup2 (pipe_command[1], STDOUT_FILENO);
    close (pipe_graph[0]), close (pipe_graph[1]);
    close (pipe_command[0]), close (pipe_command[1]);
    setsid ();
    execlp (name, name, 0);
    perror (name);
    exit (0);
  }
  else if (pid < 0) {
    perror ("fork");
    close (pipe_graph[0]), close (pipe_graph[1]);
    close (pipe_command[0]), close (pipe_command[1]);
    return 0;
  }
#else
# error "unsupported operating system"
#endif
  close (pipe_graph[0]), close (pipe_command[1]);

  if (!(visf = fdopen (pipe_command[0], "r")) ||
      !(graphf = fdopen (pipe_graph[1], "w"))) {
    perror ("fdopen");
    if (visf)
      fclose (visf);
    else
      close (visfd);
    close (pipe_graph[1]);
    return 0;
  }

  theDotty.myOut = new class Printer;
  theDotty.myOut->setWidth (0);
  theDotty.myOut->setOutput (graphf);
  visfd = pipe_command[0];
  vispid = pid;
  return &theDotty;
}

void
Dotty::stopVisual ()
{
  if (!vispid)
    return;
  assert (visf && graphf && visfd && myOut);
  vispid = 0;
  fclose (visf);
  visf = 0;
  visfd = 0;
}

void
Dotty::clearVisual ()
{
  if (!graphf && !myOut)
    return;
  assert (graphf && myOut);
  fclose (graphf);
  graphf = 0;
  delete myOut;
  myOut = 0;
}

void
Dotty::setParams (class Graph& graph,
		  class ComponentGraph* cgraph,
#ifdef EXPR_COMPILE
		  const class Compilation* compilation,
#endif // EXPR_COMPILE
		  const class Printer& printer)
{
  myGraph = &graph;
  myComponentGraph = cgraph;
#ifdef EXPR_COMPILE
  myCompilation = compilation;
#endif // EXPR_COMPILE
  myPrinter = &printer;
}

/** Backslashify a string
 * @param buf		the string buffer
 * @param record	flag: escape characters for records
 */
static void
backslashify (class StringBuffer& buf,
	      bool record)
{
  /** The characters that must be escaped */
  static const char esc1[] = "\\\"<>{|}", esc2[] = "\\\"";
  unsigned length = buf.getLength ();
  /** Number of required escape characters */
  unsigned numEscapes = 0;
  unsigned i = length;
  if (record) {
    for (const char* s = buf.getString (); i--; s++)
      if (memchr (esc1, *s, sizeof esc1))
	numEscapes++;
  }
  else {
    for (const char* s = buf.getString (); i--; s++)
      if (memchr (esc2, *s, sizeof esc2))
	numEscapes++;
  }
  if (!numEscapes)
    return;
  char* str = buf.extend (numEscapes);
  memmove (str + numEscapes, str, length);
  length += numEscapes;
  for (i = 0; i < length; i++) {
    register char c = str[i + numEscapes];
    if (record
	? memchr (esc1, c, sizeof esc1)
	: memchr (esc2, c, sizeof esc2)) {
      numEscapes--;
      str[i++] = '\\';
    }
    str[i] = c;
  }
}

/** Display the graph prologue
 * @param adding	flag: add to existing graph
 */
static void
displayPrologue (bool adding)
{
  fputs (adding ? "add" : "new", graphf);
  fputs ("();\n"
	 "digraph g{\n"
	 "nodesep=.05;\n"
	 "node[shape=record];\n", graphf);
}

/** Display the graph epilogue */
static void
displayEpilogue ()
{
  fputs ("}\n", graphf);
  fflush (graphf);
}

/** Display a marking
 * @param printer	the output stream
 * @param m		the marking to be displayed
 */
static void
displayMarking (const class Printer& printer,
		const class GlobalMarking& m)
{
  /** output buffer needed for backslashifying special characters */
  class StringBuffer out;
  class Printer pr;
  pr.setOutput (&out);
  unsigned i;
  for (i = 0; i < m.getSize (); i++) {
    if (m[i].empty () || m[i].getPlace ()->isSuppressed ())
      continue;
    pr.print (m[i].getPlace ()->getName ());
    backslashify (out, true);
    out.append (':', 1);
    printer.printRaw (out.getString ());
    out.create (0);
    bool comma = false;
    for (PlaceMarking::const_iterator t = m[i].begin ();
	 t != m[i].end (); t++) {
      if (!t->second)
	continue;
      if (comma)
	printer.delimiter (',');
      comma = true;
      if (t->second > 1)
	printer.print (t->second), printer.delimiter ('#');
      t->first->display (pr);
      backslashify (out, true);
      printer.printRaw (out.getString ());
      out.create (0);
    }
    printer.printRaw ("\\l");
  }
}

void
Dotty::dump (bool adding) const
{
  fputs (adding ? "add" : "new", graphf);
  fputs ("();\n"
	 "digraph g{\n"
	 "nodesep=.05;\n", graphf);

  unsigned i;
  const class Net& net = myGraph->getNet ();
  class StringBuffer out;
  class Printer pr;
  pr.setOutput (&out);

  for (i = net.getNumPlaces (); i--; ) {
    pr.print (net.getPlace (i).getName ());
    backslashify (out, false);
    fprintf (graphf, "%u[label=\"", i);
    fwrite (out.getString (), out.getLength (), 1, graphf);
    out.create (0);
    fputs ("\\n", graphf);
    net.getPlace (i).getType ().display (pr);
    backslashify (out, false);
    fwrite (out.getString (), out.getLength (), 1, graphf);
    out.create (0);
    fputs ("\"];\n", graphf);
  }
  fputs ("node[shape=box];\n", graphf);
  for (i = net.getNumTransitions (); i--; ) {
    const class Transition& tr = net.getTransition (i);
    pr.print (tr.getName ());
    backslashify (out, false);
    fprintf (graphf, "t%u[label=\"", i);
    fwrite (out.getString (), out.getLength (), 1, graphf);
    out.create (0);
    for (Transition::GateList::const_iterator g = tr.beginG ();
	 g != tr.endG (); g++) {
      fputs ("\\n", graphf);
      (*g)->display (pr);
      backslashify (out, false);
      fwrite (out.getString (), out.getLength (), 1, graphf);
      out.create (0);
    }
    fputs ("\"];\n", graphf);
    unsigned j;
    for (j = tr.getNumInputs (); j--; ) {
      const class Arc& arc = tr.getInput (j);
      fprintf (graphf, "%u->t%u[label=\"", arc.getPlace ().getIndex (), i);
      arc.getExpr ().display (pr);
      backslashify (out, false);
      fwrite (out.getString (), out.getLength (), 1, graphf);
      out.create (0);
      fputs ("\"];\n", graphf);
    }
    for (j = tr.getNumOutputs (); j--; ) {
      const class Arc& arc = tr.getOutput (j);
      fprintf (graphf, "t%u->%u[label=\"", i, arc.getPlace ().getIndex ());
      arc.getExpr ().display (pr);
      backslashify (out, false);
      fwrite (out.getString (), out.getLength (), 1, graphf);
      out.create (0);
      fputs ("\"];\n", graphf);
    }
  }
  fputs ("}\n", graphf);
  fflush (graphf);
}

/** Display a state
 * @param printer	the output stream
 * @param graph		the reachability graph
 * @param cgraph	the component graph (optional)
 * @param state		index number of the state
 */
static void
displayState (const class Printer& printer,
	      const class Graph& graph,
	      const class ComponentGraph* cgraph,
	      card_t state)
{
  class GlobalMarking* m = graph.fetchState (state);
  card_t* succ = graph.getSuccessors (state);
  assert (!succ || *succ);
  card_t numSucc = succ ? *succ : 0;
  bool isProcessed = graph.isProcessed (state);
  delete[] succ;
  printer.printRaw ("\"");
  printer.printState (state);
  printer.printRaw ("\"[");
  if (graph.isErroneous (state))
    printer.printRaw (isProcessed
		      ? "style=filled,color=coral,fontcolor=yellow,"
		      : "style=filled,color=firebrick,fontcolor=yellow,");
  else if (!numSucc)
    printer.printRaw (isProcessed
		      ? "style=filled,color=black,fontcolor=white,"
		      : "style=filled,color=slategray,fontcolor=white,");
  printer.printRaw ("label=\"{");
  printer.printState (state);
  if (isProcessed) {
    printer.delimiter ('|');
    printer.print (numSucc);
    if (card_t numPred = graph.getNumPredecessors (state)) {
      printer.delimiter ('(');
      printer.print (numPred);
      printer.delimiter (')');
    }
  }
  if (cgraph) {
    printer.printRaw ("|");
    printer.printComponent (cgraph->getComponent (state));
  }
  printer.printRaw ("}|");
  ::displayMarking (printer, *m);
  printer.printRaw ("\"];");
  printer.finish ();
  delete m;
}

void
Dotty::show (card_t state, bool adding) const
{
  ::displayPrologue (adding);
  ::displayState (*myOut, *myGraph, myComponentGraph, state);
  ::displayEpilogue ();
}

/** Display an edge in the reachability graph
 * @param printer	the output stream
 * @param transition	the transition
 * @param valuation	the valuation
 * @param source	the source state
 * @param dest		the destination state
 * @param index		index number of the edge
 */
static void
displayEdge (const class Printer& printer,
	     const class Transition& transition,
	     const class Valuation& valuation,
	     card_t source,
	     card_t dest,
	     unsigned index)
{
  assert (valuation.isOK ());
  class StringBuffer out;
  class Printer pr;
  pr.setOutput (&out);
  printer.printRaw ("\"");
  printer.printState (source);
  printer.printRaw ("\"->\"");
  printer.printState (dest);
  printer.printRaw ("\"[id=");
  printer.print (index);
  printer.printRaw (",label=\"");
  pr.print (transition.getName ());
  ::backslashify (out, false);
  out.append ("\\n");
  printer.printRaw (out.getString ());
  out.create (0);
  for (Valuation::const_iterator i = valuation.begin ();
       i != valuation.end (); i++) {
    pr.print (i->first->getName ());
    pr.delimiter (':');
    i->second->display (pr);
    backslashify (out, false);
    out.append ("\\l");
    printer.printRaw (out.getString ());
    out.create (0);
  }
  printer.printRaw ("\"];");
  printer.finish ();
}

void
Dotty::succ (card_t state, bool adding) const
{
#ifdef EXPR_COMPILE
  if (myCompilation)
    myGraph->getNet ().analyze (*myCompilation,
				*myPrinter, *myGraph, state, Net::Single);
  else
#endif // EXPR_COMPILE
    myGraph->getNet ().analyze (*myPrinter, *myGraph, state, Net::Single);
  word_t* data;
  card_t* succ = myGraph->getSuccessors (state, &data);
  assert (!succ || *succ);
  ::displayPrologue (adding);
  ::displayState (*myOut, *myGraph, myComponentGraph, state);
  if (succ) {
    class BitUnpacker buf (data);
    for (card_t i = 1; i <= *succ; i++) {
      const class Transition* transition;
      class Valuation v;
      myGraph->getSuccessor (buf, transition, v);
      ::displayEdge (*myOut, *transition, v, state, succ[i], i);
      ::displayState (*myOut, *myGraph, myComponentGraph, succ[i]);
    }
    delete[] data;
    delete[] succ;
  }
  ::displayEpilogue ();
}

void
Dotty::pred (card_t state, bool adding) const
{
  ::displayPrologue (adding);
  Graph::fpos_t pos = myGraph->getPredecessors (state);
  for (card_t oldsource = CARD_T_MAX; pos != FPOS_NONE; ) {
    card_t source = myGraph->getPredecessor (pos);
    if (oldsource == source)
      continue; /* several arcs from the same state */
    ::displayState (*myOut, *myGraph, myComponentGraph, oldsource = source);
    word_t* data;
    card_t* succ = myGraph->getSuccessors (source, &data);
    assert (succ && *succ);
    if (succ) {
      class BitUnpacker buf (data);
      for (card_t i = 1; i <= *succ; i++) {
	const class Transition* transition;
	class Valuation v;
	myGraph->getSuccessor (buf, transition, v);
	if (succ[i] != state)
	  continue;
	::displayEdge (*myOut, *transition, v, source, state, i);
      }
      delete[] data;
      delete[] succ;
    }
  }
  ::displayState (*myOut, *myGraph, myComponentGraph, state);
  ::displayEpilogue ();
}

/** Display a strongly connected component in a component graph
 * @param printer	the output stream
 * @param cgraph	the component graph
 * @param comp		the component number
 */
static void
displayComponent (const class Printer& printer,
		  const class ComponentGraph& cgraph,
		  card_t comp)
{
  card_t* s = cgraph.getStates (comp);
  assert (s && *s);
  card_t numPred = cgraph.getNumPred (comp);
  card_t numSucc = cgraph.getNumSucc (comp);
  printer.printRaw ("\"");
  printer.printComponent (comp);
  printer.printRaw ("\"[label=\"");
  printer.printComponent (comp);
  printer.printNumber ("|", *s);
  printer.printNumber ("|", numSucc);
  printer.printNumber ("(", numPred);
  printer.printRaw (")\"]");
  printer.finish ();
  delete[] s;
}

/** Display an edge in the strongly connected component graph
 * @param printer	the output stream
 * @param source	the source component
 * @param dest		the destination component
 */
static void
displayComponentEdge (const class Printer& printer,
		      card_t source, card_t dest)
{
  printer.printRaw ("\"");
  printer.printComponent (source);
  printer.printRaw ("\"->\"");
  printer.printComponent (dest);
  printer.printRaw ("\"");
  printer.finish ();
}

void
Dotty::components (bool adding) const
{
  ::displayPrologue (adding);
  for (card_t comp = myComponentGraph->size (); comp--; ) {
    ::displayComponent (*myOut, *myComponentGraph, comp);
    if (card_t* s = myComponentGraph->getSucc (comp)) {
      assert (*s > 0);
      for (card_t i = *s; i; i--)
	::displayComponentEdge (*myOut, comp, s[i]);
      delete[] s;
    }
  }
  ::displayEpilogue ();
}

void
Dotty::component (card_t comp,
		  const class Expression* cond,
		  bool adding) const
{
  card_t* states = myComponentGraph->getStates (comp);
  assert (states && *states);
  ::displayPrologue (adding);
  card_t state;
  if (cond)
    for (state = *states; state; state--)
      if (!myGraph->eval (states[state], cond))
	memmove (states + state, states + state + 1,
		 (*states - state) * sizeof *states),
	  (*states)--;
  for (state = *states; state; state--) {
    ::displayState (*myOut, *myGraph, myComponentGraph, states[state]);
    word_t* data;
    if (card_t* succ = myGraph->getSuccessors (states[state], &data)) {
      class BitUnpacker buf (data);
      for (card_t i = 1; i <= *succ; i++) {
	const class Transition* transition;
	class Valuation v;
	myGraph->getSuccessor (buf, transition, v);
	register card_t dest = succ[i];
	for (card_t* s = states + *states; s > states; s--)
	  if (dest == *s)
	    goto inComponent;
	continue;
      inComponent:
	::displayEdge (*myOut, *transition, v, states[state], dest, i);
      }
      delete[] data;
      delete[] succ;
    }
  }
  delete[] states;
  ::displayEpilogue ();
}

void
Dotty::comp_succ (card_t comp, bool adding) const
{
  ::displayPrologue (adding);
  ::displayComponent (*myOut, *myComponentGraph, comp);
  if (card_t* s = myComponentGraph->getSucc (comp)) {
    assert (*s > 0);
    for (card_t i = *s; i; i--) {
      ::displayComponent (*myOut, *myComponentGraph, s[i]);
      ::displayComponentEdge (*myOut, comp, s[i]);
    }
    delete[] s;
  }
  ::displayEpilogue ();
}

void
Dotty::comp_pred (card_t comp, bool adding) const
{
  ::displayPrologue (adding);
  ::displayComponent (*myOut, *myComponentGraph, comp);
  if (card_t* s = myComponentGraph->getPred (comp)) {
    assert (*s > 0);
    for (card_t i = *s; i; i--) {
      ::displayComponent (*myOut, *myComponentGraph, s[i]);
      ::displayComponentEdge (*myOut, s[i], comp);
    }
    delete[] s;
  }
  ::displayEpilogue ();
}

void
Dotty::show (const card_t* path, bool adding, bool reverse) const
{
  assert (!!path);
  ::displayPrologue (adding);
  for (const card_t* p = path + *path; p > path; p--) {
    ::displayState (*myOut, *myGraph, myComponentGraph, *p);
    if (const card_t* next = reverse
	? (p < path + *path ? p + 1 : 0)
	: (p > path + 1 ? p - 1 : 0)) {
      word_t* data;
      card_t* succ = myGraph->getSuccessors (*p, &data);
      assert (!succ || *succ);
      if (succ) {
	class BitUnpacker buf (data);
	for (card_t i = 1; i <= *succ; i++) {
	  const class Transition* transition;
	  class Valuation v;
	  myGraph->getSuccessor (buf, transition, v);
	  if (succ[i] != *next)
	    continue;
	  ::displayEdge (*myOut, *transition, v, *p, *next, i);
	}
	delete[] data;
	delete[] succ;
      }
    }
  }
  ::displayEpilogue ();
}
